package ctbrec.ui.tabs.recorded;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import ctbrec.Config;
import ctbrec.Model;
import ctbrec.ModelGroup;
import ctbrec.image.LocalPortraitStore;
import ctbrec.image.PortraitStore;
import ctbrec.image.RemotePortraitStore;
import ctbrec.io.HttpClient;
import ctbrec.io.json.ObjectMapperFactory;
import ctbrec.io.json.dto.ModelDto;
import ctbrec.io.json.mapper.MappingException;
import ctbrec.io.json.mapper.ModelMapper;
import ctbrec.sites.Site;
import ctbrec.sites.SiteUtil;
import lombok.extern.slf4j.Slf4j;
import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONWriter;
import org.mapstruct.factory.Mappers;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.*;
import java.util.stream.Collectors;

@Slf4j
public class ModelImportExport {


    enum ExportIncludes {
        NOTES,
        GROUPS,
        PORTRAITS
    }

    private static final ObjectMapper mapper = ObjectMapperFactory.getMapper();

    record ExportOptions(Set<ExportIncludes> includes, File targetFile) {
    }

    private ModelImportExport() {
    }

    public static void exportTo(List<Model> models, Set<ModelGroup> groups, Config config, ExportOptions exportOptions) throws IOException {
        StringBuilder sb = new StringBuilder();
        JSONWriter writer = new JSONWriter(sb);

        writer.object();
        writer.key("models");
        writer.array();
        var modelArray = models.stream()
                .map(Mappers.getMapper(ModelMapper.class)::toDto)
                .map(dto -> {
                    try {
                        return mapper.writeValueAsString(dto);
                    } catch (JsonProcessingException e) {
                        log.error("Error while serializing model {}", dto);
                        throw new MappingException(e);
                    }
                })
                .collect(Collectors.joining(","));
        sb.append(modelArray);
        writer.endArray();
        if (exportOptions.includes().contains(ExportIncludes.NOTES)) {
            writer.key("notes");
            writer.value(config.getSettings().modelNotes);
        }
        if (exportOptions.includes().contains(ExportIncludes.GROUPS)) {
            writer.key("groups");
            writer.array();
            var groupArray = groups
                    .stream().map(grp -> {
                        try {
                            return mapper.writeValueAsString(grp);
                        } catch (JsonProcessingException e) {
                            log.error("Error while serializing model group {}", grp);
                            throw new MappingException(e);
                        }
                    })
                    .collect(Collectors.joining(","));
            sb.append(groupArray);
            writer.endArray();
        }
        if (exportOptions.includes().contains(ExportIncludes.PORTRAITS)) {
            var portraits = config.getSettings().modelPortraits;
            PortraitStore portraitLoader;
            if (config.getSettings().localRecording) {
                portraitLoader = new LocalPortraitStore(config);
            } else {
                var httpClient = new HttpClient("camrec", config) {
                    @Override
                    public boolean login() {
                        return false;
                    }
                };
                portraitLoader = new RemotePortraitStore(httpClient, config);
            }
            if (portraits != null && !portraits.isEmpty()) {
                writer.key("portraits");
                writer.array();
                for (Map.Entry<String, String> entry : config.getSettings().modelPortraits.entrySet()) {
                    String modelUrl = entry.getKey();
                    String portraitId = entry.getValue();
                    Optional<byte[]> portrait = portraitLoader.loadModelPortraitByModelUrl(modelUrl);
                    if (portrait.isPresent()) {
                        writer.object();
                        writer.key("url").value(modelUrl);
                        writer.key("id").value(portraitId);
                        writer.key("data").value(Base64.getEncoder().encodeToString(portrait.get()));
                        writer.endObject();
                    }
                }
                writer.endArray();
            }
        }
        writer.endObject();
        Files.writeString(exportOptions.targetFile().toPath(), new JSONObject(sb.toString()).toString(2));
    }

    public static List<Model> importFrom(File target, List<Site> sites, Config config) throws IOException {
        JSONObject json = new JSONObject(Files.readString(target.toPath(), StandardCharsets.UTF_8));
        List<Model> models = readModels(json.getJSONArray("models"));
        models.forEach(m -> SiteUtil.getSiteForModel(sites, m).ifPresent(m::setSite));
        if (json.has("notes")) {
            importNotes(json.getJSONObject("notes"), config);
        }
        if (json.has("groups")) {
            importGroups(json.getJSONArray("groups"), config);
        }
        if (json.has("portraits")) {
            importPortraits(json.getJSONArray("portraits"), config);
        }
        return models;
    }

    private static void importPortraits(JSONArray portraits, Config config) throws IOException {
        PortraitStore portraitStore;
        if (config.getSettings().localRecording) {
            portraitStore = new LocalPortraitStore(config);
        } else {
            var httpClient = new HttpClient("camrec", config) {
                @Override
                public boolean login() {
                    return false;
                }
            };
            portraitStore = new RemotePortraitStore(httpClient, config);
        }
        for (int i = 0; i < portraits.length(); i++) {
            JSONObject portrait = portraits.getJSONObject(i);
            String url = portrait.getString("url");
            String id = portrait.getString("id");
            String dataBase64 = portrait.getString("data");
            portraitStore.writePortrait(url, Base64.getDecoder().decode(dataBase64));
        }
    }


    private static void importGroups(JSONArray groups, Config config) {
        for (int i = 0; i < groups.length(); i++) {
            JSONObject group = groups.getJSONObject(i);
            try {
                ModelGroup modelGroup = mapper.readValue(group.toString(), ModelGroup.class);
                config.getSettings().modelGroups.add(modelGroup);
            } catch (JsonProcessingException e) {
                log.error("Error while deserializing model group {}", group);
            }
        }
    }

    private static void importNotes(JSONObject notes, Config config) {
        var modelNotes = new HashMap<String, String>();
        JSONArray urls = notes.names();
        for (int i = 0; i < urls.length(); i++) {
            String url = urls.getString(i);
            String note = notes.getString(url);
            modelNotes.put(url, note);
        }
        config.getSettings().modelNotes.putAll(modelNotes);
    }

    private static List<Model> readModels(JSONArray models) {
        List<Model> result = new LinkedList<>();
        for (int i = 0; i < models.length(); i++) {
            JSONObject model = models.getJSONObject(i);
            try {
                ModelDto dto = mapper.readValue(model.toString(), ModelDto.class);
                result.add(Mappers.getMapper(ModelMapper.class).toModel(dto));
            } catch (Exception e) {
                log.error("Error while deserializing model {}", model.toString(2), e);
            }
        }
        return result;
    }
}
